def addition():
    """
    Turn the sequence of numbers into a sum

    + or - ?
    """


def multiply():
    """

    * or /

    """


def whatsbigger():
    """
    Given two numbers, determine which one bigger
    """


earef
