def reverse():
    """
    Take a sequence and reverse it
    """
    pass


def find_repeat():
    """
    Take a sequence, find a pattern in it and repeat that pattern
    """
    pass


def sort():
    """
    Sort a sequence
    """

def map_one_to_one():
    """
    Given a sequence, learn to map it to another thing

    Mappings can be many to one, one to many
    """


def map_one_to_many():


def map_many_to_one():


def map_many_to_many():
    """
    will need to Define rules for context leading to a result


    """


def numerical():
    """
    Test whether the model can do certain sums
    """


def summarisation():





